/* Copyright (c) 2005 Dirk Jagdmann <doj@cubic.org>

This software is provided 'as-is', without any express or implied
warranty. In no event will the authors be held liable for any damages
arising from the use of this software.

Permission is granted to anyone to use this software for any purpose,
including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

1. The origin of this software must not be misrepresented; you must
not claim that you wrote the original software. If you use this
software in a product, an acknowledgment in the product documentation
would be appreciated but is not required.

2. Altered source versions must be plainly marked as such, and must
not be misrepresented as being the original software.

3. This notice may not be removed or altered from any source
distribution. */

/* $Header: /code/doj/wav.h,v 1.16 2005/06/23 16:24:35 doj Exp $ */

/* RIFF/WAVE files are always little-endian */

/* TODO:
   + make c++ read/write access to variables portable to big/little endian machines
   + in WAVload() make a real parser for different chunks
*/

#ifndef WAV__H
#define WAV__H

#include <string.h>
#include <stdint.h>

#ifndef NONEWWAVE
/* WAVE form wFormatTag IDs */
#define  WAVE_FORMAT_UNKNOWN                    0x0000 /* Microsoft Corporation */
#define  WAVE_FORMAT_PCM                        0x0001
#define  WAVE_FORMAT_ADPCM                      0x0002 /* Microsoft Corporation */
#define  WAVE_FORMAT_IEEE_FLOAT                 0x0003 /* Microsoft Corporation */
#define  WAVE_FORMAT_VSELP                      0x0004 /* Compaq Computer Corp. */
#define  WAVE_FORMAT_IBM_CVSD                   0x0005 /* IBM Corporation */
#define  WAVE_FORMAT_ALAW                       0x0006 /* Microsoft Corporation */
#define  WAVE_FORMAT_MULAW                      0x0007 /* Microsoft Corporation */
#define  WAVE_FORMAT_DTS                        0x0008 /* Microsoft Corporation */
#define  WAVE_FORMAT_DRM                        0x0009 /* Microsoft Corporation */
#define  WAVE_FORMAT_OKI_ADPCM                  0x0010 /* OKI */
#define  WAVE_FORMAT_DVI_ADPCM                  0x0011 /* Intel Corporation */
#define  WAVE_FORMAT_IMA_ADPCM                  (WAVE_FORMAT_DVI_ADPCM) /*  Intel Corporation */
#define  WAVE_FORMAT_MEDIASPACE_ADPCM           0x0012 /* Videologic */
#define  WAVE_FORMAT_SIERRA_ADPCM               0x0013 /* Sierra Semiconductor Corp */
#define  WAVE_FORMAT_G723_ADPCM                 0x0014 /* Antex Electronics Corporation */
#define  WAVE_FORMAT_DIGISTD                    0x0015 /* DSP Solutions, Inc. */
#define  WAVE_FORMAT_DIGIFIX                    0x0016 /* DSP Solutions, Inc. */
#define  WAVE_FORMAT_DIALOGIC_OKI_ADPCM         0x0017 /* Dialogic Corporation */
#define  WAVE_FORMAT_MEDIAVISION_ADPCM          0x0018 /* Media Vision, Inc. */
#define  WAVE_FORMAT_CU_CODEC                   0x0019 /* Hewlett-Packard Company */
#define  WAVE_FORMAT_YAMAHA_ADPCM               0x0020 /* Yamaha Corporation of America */
#define  WAVE_FORMAT_SONARC                     0x0021 /* Speech Compression */
#define  WAVE_FORMAT_DSPGROUP_TRUESPEECH        0x0022 /* DSP Group, Inc */
#define  WAVE_FORMAT_ECHOSC1                    0x0023 /* Echo Speech Corporation */
#define  WAVE_FORMAT_AUDIOFILE_AF36             0x0024 /* Virtual Music, Inc. */
#define  WAVE_FORMAT_APTX                       0x0025 /* Audio Processing Technology */
#define  WAVE_FORMAT_AUDIOFILE_AF10             0x0026 /* Virtual Music, Inc. */
#define  WAVE_FORMAT_PROSODY_1612               0x0027 /* Aculab plc */
#define  WAVE_FORMAT_LRC                        0x0028 /* Merging Technologies S.A. */
#define  WAVE_FORMAT_DOLBY_AC2                  0x0030 /* Dolby Laboratories */
#define  WAVE_FORMAT_GSM610                     0x0031 /* Microsoft Corporation */
#define  WAVE_FORMAT_MSNAUDIO                   0x0032 /* Microsoft Corporation */
#define  WAVE_FORMAT_ANTEX_ADPCME               0x0033 /* Antex Electronics Corporation */
#define  WAVE_FORMAT_CONTROL_RES_VQLPC          0x0034 /* Control Resources Limited */
#define  WAVE_FORMAT_DIGIREAL                   0x0035 /* DSP Solutions, Inc. */
#define  WAVE_FORMAT_DIGIADPCM                  0x0036 /* DSP Solutions, Inc. */
#define  WAVE_FORMAT_CONTROL_RES_CR10           0x0037 /* Control Resources Limited */
#define  WAVE_FORMAT_NMS_VBXADPCM               0x0038 /* Natural MicroSystems */
#define  WAVE_FORMAT_CS_IMAADPCM                0x0039 /* Crystal Semiconductor IMA ADPCM */
#define  WAVE_FORMAT_ECHOSC3                    0x003A /* Echo Speech Corporation */
#define  WAVE_FORMAT_ROCKWELL_ADPCM             0x003B /* Rockwell International */
#define  WAVE_FORMAT_ROCKWELL_DIGITALK          0x003C /* Rockwell International */
#define  WAVE_FORMAT_XEBEC                      0x003D /* Xebec Multimedia Solutions Limited */
#define  WAVE_FORMAT_G721_ADPCM                 0x0040 /* Antex Electronics Corporation */
#define  WAVE_FORMAT_G728_CELP                  0x0041 /* Antex Electronics Corporation */
#define  WAVE_FORMAT_MSG723                     0x0042 /* Microsoft Corporation */
#define  WAVE_FORMAT_MPEG                       0x0050 /* Microsoft Corporation */
#define  WAVE_FORMAT_RT24                       0x0052 /* InSoft, Inc. */
#define  WAVE_FORMAT_PAC                        0x0053 /* InSoft, Inc. */
#define  WAVE_FORMAT_MPEGLAYER3                 0x0055 /* ISO/MPEG Layer3 Format Tag */
#define  WAVE_FORMAT_LUCENT_G723                0x0059 /* Lucent Technologies */
#define  WAVE_FORMAT_CIRRUS                     0x0060 /* Cirrus Logic */
#define  WAVE_FORMAT_ESPCM                      0x0061 /* ESS Technology */
#define  WAVE_FORMAT_VOXWARE                    0x0062 /* Voxware Inc */
#define  WAVE_FORMAT_CANOPUS_ATRAC              0x0063 /* Canopus, co., Ltd. */
#define  WAVE_FORMAT_G726_ADPCM                 0x0064 /* APICOM */
#define  WAVE_FORMAT_G722_ADPCM                 0x0065 /* APICOM */
#define  WAVE_FORMAT_DSAT_DISPLAY               0x0067 /* Microsoft Corporation */
#define  WAVE_FORMAT_VOXWARE_BYTE_ALIGNED       0x0069 /* Voxware Inc */
#define  WAVE_FORMAT_VOXWARE_AC8                0x0070 /* Voxware Inc */
#define  WAVE_FORMAT_VOXWARE_AC10               0x0071 /* Voxware Inc */
#define  WAVE_FORMAT_VOXWARE_AC16               0x0072 /* Voxware Inc */
#define  WAVE_FORMAT_VOXWARE_AC20               0x0073 /* Voxware Inc */
#define  WAVE_FORMAT_VOXWARE_RT24               0x0074 /* Voxware Inc */
#define  WAVE_FORMAT_VOXWARE_RT29               0x0075 /* Voxware Inc */
#define  WAVE_FORMAT_VOXWARE_RT29HW             0x0076 /* Voxware Inc */
#define  WAVE_FORMAT_VOXWARE_VR12               0x0077 /* Voxware Inc */
#define  WAVE_FORMAT_VOXWARE_VR18               0x0078 /* Voxware Inc */
#define  WAVE_FORMAT_VOXWARE_TQ40               0x0079 /* Voxware Inc */
#define  WAVE_FORMAT_SOFTSOUND                  0x0080 /* Softsound, Ltd. */
#define  WAVE_FORMAT_VOXWARE_TQ60               0x0081 /* Voxware Inc */
#define  WAVE_FORMAT_MSRT24                     0x0082 /* Microsoft Corporation */
#define  WAVE_FORMAT_G729A                      0x0083 /* AT&T Labs, Inc. */
#define  WAVE_FORMAT_MVI_MVI2                   0x0084 /* Motion Pixels */
#define  WAVE_FORMAT_DF_G726                    0x0085 /* DataFusion Systems (Pty) (Ltd) */
#define  WAVE_FORMAT_DF_GSM610                  0x0086 /* DataFusion Systems (Pty) (Ltd) */
#define  WAVE_FORMAT_ISIAUDIO                   0x0088 /* Iterated Systems, Inc. */
#define  WAVE_FORMAT_ONLIVE                     0x0089 /* OnLive! Technologies, Inc. */
#define  WAVE_FORMAT_SBC24                      0x0091 /* Siemens Business Communications Sys */
#define  WAVE_FORMAT_DOLBY_AC3_SPDIF            0x0092 /* Sonic Foundry */
#define  WAVE_FORMAT_MEDIASONIC_G723            0x0093 /* MediaSonic */
#define  WAVE_FORMAT_PROSODY_8KBPS              0x0094 /* Aculab plc */
#define  WAVE_FORMAT_ZYXEL_ADPCM                0x0097 /* ZyXEL Communications, Inc. */
#define  WAVE_FORMAT_PHILIPS_LPCBB              0x0098 /* Philips Speech Processing */
#define  WAVE_FORMAT_PACKED                     0x0099 /* Studer Professional Audio AG */
#define  WAVE_FORMAT_MALDEN_PHONYTALK           0x00A0 /* Malden Electronics Ltd. */
#define  WAVE_FORMAT_RHETOREX_ADPCM             0x0100 /* Rhetorex Inc. */
#define  WAVE_FORMAT_IRAT                       0x0101 /* BeCubed Software Inc. */
#define  WAVE_FORMAT_VIVO_G723                  0x0111 /* Vivo Software */
#define  WAVE_FORMAT_VIVO_SIREN                 0x0112 /* Vivo Software */
#define  WAVE_FORMAT_DIGITAL_G723               0x0123 /* Digital Equipment Corporation */
#define  WAVE_FORMAT_SANYO_LD_ADPCM             0x0125 /* Sanyo Electric Co., Ltd. */
#define  WAVE_FORMAT_SIPROLAB_ACEPLNET          0x0130 /* Sipro Lab Telecom Inc. */
#define  WAVE_FORMAT_SIPROLAB_ACELP4800         0x0131 /* Sipro Lab Telecom Inc. */
#define  WAVE_FORMAT_SIPROLAB_ACELP8V3          0x0132 /* Sipro Lab Telecom Inc. */
#define  WAVE_FORMAT_SIPROLAB_G729              0x0133 /* Sipro Lab Telecom Inc. */
#define  WAVE_FORMAT_SIPROLAB_G729A             0x0134 /* Sipro Lab Telecom Inc. */
#define  WAVE_FORMAT_SIPROLAB_KELVIN            0x0135 /* Sipro Lab Telecom Inc. */
#define  WAVE_FORMAT_G726ADPCM                  0x0140 /* Dictaphone Corporation */
#define  WAVE_FORMAT_QUALCOMM_PUREVOICE         0x0150 /* Qualcomm, Inc. */
#define  WAVE_FORMAT_QUALCOMM_HALFRATE          0x0151 /* Qualcomm, Inc. */
#define  WAVE_FORMAT_TUBGSM                     0x0155 /* Ring Zero Systems, Inc. */
#define  WAVE_FORMAT_MSAUDIO1                   0x0160 /* Microsoft Corporation */
#define  WAVE_FORMAT_UNISYS_NAP_ADPCM           0x0170 /* Unisys Corp. */
#define  WAVE_FORMAT_UNISYS_NAP_ULAW            0x0171 /* Unisys Corp. */
#define  WAVE_FORMAT_UNISYS_NAP_ALAW            0x0172 /* Unisys Corp. */
#define  WAVE_FORMAT_UNISYS_NAP_16K             0x0173 /* Unisys Corp. */
#define  WAVE_FORMAT_CREATIVE_ADPCM             0x0200 /* Creative Labs, Inc */
#define  WAVE_FORMAT_CREATIVE_FASTSPEECH8       0x0202 /* Creative Labs, Inc */
#define  WAVE_FORMAT_CREATIVE_FASTSPEECH10      0x0203 /* Creative Labs, Inc */
#define  WAVE_FORMAT_UHER_ADPCM                 0x0210 /* UHER informatic GmbH */
#define  WAVE_FORMAT_QUARTERDECK                0x0220 /* Quarterdeck Corporation */
#define  WAVE_FORMAT_ILINK_VC                   0x0230 /* I-link Worldwide */
#define  WAVE_FORMAT_RAW_SPORT                  0x0240 /* Aureal Semiconductor */
#define  WAVE_FORMAT_ESST_AC3                   0x0241 /* ESS Technology, Inc. */
#define  WAVE_FORMAT_IPI_HSX                    0x0250 /* Interactive Products, Inc. */
#define  WAVE_FORMAT_IPI_RPELP                  0x0251 /* Interactive Products, Inc. */
#define  WAVE_FORMAT_CS2                        0x0260 /* Consistent Software */
#define  WAVE_FORMAT_SONY_SCX                   0x0270 /* Sony Corp. */
#define  WAVE_FORMAT_FM_TOWNS_SND               0x0300 /* Fujitsu Corp. */
#define  WAVE_FORMAT_BTV_DIGITAL                0x0400 /* Brooktree Corporation */
#define  WAVE_FORMAT_QDESIGN_MUSIC              0x0450 /* QDesign Corporation */
#define  WAVE_FORMAT_VME_VMPCM                  0x0680 /* AT&T Labs, Inc. */
#define  WAVE_FORMAT_TPC                        0x0681 /* AT&T Labs, Inc. */
#define  WAVE_FORMAT_OLIGSM                     0x1000 /* Ing C. Olivetti & C., S.p.A. */
#define  WAVE_FORMAT_OLIADPCM                   0x1001 /* Ing C. Olivetti & C., S.p.A. */
#define  WAVE_FORMAT_OLICELP                    0x1002 /* Ing C. Olivetti & C., S.p.A. */
#define  WAVE_FORMAT_OLISBC                     0x1003 /* Ing C. Olivetti & C., S.p.A. */
#define  WAVE_FORMAT_OLIOPR                     0x1004 /* Ing C. Olivetti & C., S.p.A. */
#define  WAVE_FORMAT_LH_CODEC                   0x1100 /* Lernout & Hauspie */
#define  WAVE_FORMAT_NORRIS                     0x1400 /* Norris Communications, Inc. */
#define  WAVE_FORMAT_SOUNDSPACE_MUSICOMPRESS    0x1500 /* AT&T Labs, Inc. */
#define  WAVE_FORMAT_DVM                        0x2000 /* FAST Multimedia AG */

#if !defined(WAVE_FORMAT_EXTENSIBLE)
#define  WAVE_FORMAT_EXTENSIBLE                 0xFFFE /* Microsoft */
#endif

#define WAVE_FORMAT_DEVELOPMENT         (0xFFFF)

#define NONEWWAVE
#endif /* NONEWWAVE */

#pragma pack(push, 1)

struct WAVfmt
{
  char name[4];			/* "fmt " */
  uint32_t length;		/* 0x10 */
  uint16_t type;		/* WAVE_FORMAT_... */
  uint16_t channels;		/* commonly 1,2 but can be >2 */
  uint32_t rate;
  uint32_t bytesPerSecond;	/* rate*bytesPerSample */
  union {
    uint16_t blockAlign;
    uint16_t bytesPerSample;	/* channels*((bitsPerSample+7)/8) */
  };
  uint16_t bitsPerSample;	/* 8,12,16,24,32 */
#ifdef __cplusplus
  WAVfmt() :
    length(0x10),
    type(WAVE_FORMAT_PCM),
    channels(2),
    rate(44100),
    bitsPerSample(16)
  {
    name[0]='f';
    name[1]='m';
    name[2]='t';
    name[3]=' ';
    fix();
  }
  void fix()
  {
    bytesPerSample=channels*(bitsPerSample/8);
    bytesPerSecond=rate*bytesPerSample;
  }
  bool iswav()
  {
    if(memcmp(name, "fmt ", sizeof(name))) return false;
    if(length!=0x10) return false;
    if(bytesPerSample != channels*((bitsPerSample+7)/8)) return false;
    if(bytesPerSecond != rate*bytesPerSample) return false;
    return true;
  }
#endif
};

struct WAVdata
{
  char name[4];			/* "data" */
  uint32_t length;		/* length of following sample stream = WAV.length - 0x24 = filelength - 8 - 0x24 */
#ifdef __cplusplus
  WAVdata() :
    length(0)
  {
    name[0]='d';
    name[1]='a';
    name[2]='t';
    name[3]='a';
  }
  bool iswav()
  {
    return memcmp(name, "data", sizeof(name))==0;
  }
  void* data()
  {
    return reinterpret_cast<char*>(this)+sizeof(WAVdata);
  }
#endif
};

struct WAV
{
  char name[4];			/* "RIFF" for little endian, "RIFX" for big endian files */
  uint32_t length;		/* filelength - 8 */
  char type[4];			/* "WAVE" */
  struct WAVfmt fmt;
  struct WAVdata data;
#ifdef __cplusplus
  WAV() :
    length(0x24)
  {
    name[0]='R';
    name[1]='I';
    name[2]='F';
    name[3]='F';
    type[0]='W';
    type[1]='A';
    type[2]='V';
    type[3]='E';
  }
  bool iswav()
  {
    if(memcmp(name, "RIFF", sizeof(name))) return false;
    if(memcmp(type, "WAVE", sizeof(type))) return false;
    return fmt.iswav() && data.iswav();
  }
  void setlength(int l)
  {
    length=l-8;
    data.length=length-0x24;
  }
#endif
};

#pragma pack(pop)


/* various helper functions */
#ifdef __cplusplus
#include <cstdio>
#include <iterator>
#include "file.hpp"

#ifdef USE_BOOST
#include <boost/static_assert.hpp>
#else
#define BOOST_STATIC_ASSERT(x)
#endif

namespace {

  template <typename T>
  struct WAVtraits
  {
    enum { MAGIC = WAVE_FORMAT_UNKNOWN, MAX = 0, OFFSET = 0 };
  };

  template <>
  struct WAVtraits<int8_t>
  {
    enum { MAGIC = WAVE_FORMAT_UNKNOWN, MAX = 0x7F, OFFSET = 0 };
  };

  template <>
  struct WAVtraits<uint8_t>
  {
    enum { MAGIC = WAVE_FORMAT_PCM, MAX = 0xFF, OFFSET = 0x80 };
  };

  template <>
  struct WAVtraits<int16_t>
  {
    enum { MAGIC = WAVE_FORMAT_PCM, MAX = 0x7FFF, OFFSET = 0 };
  };

  template <>
  struct WAVtraits<uint16_t>
  {
    enum { MAGIC = WAVE_FORMAT_UNKNOWN, MAX = 0xFFFF, OFFSET = 0x8000 };
  };

  template <>
  struct WAVtraits<int32_t>
  {
    enum { MAGIC = WAVE_FORMAT_PCM, MAX = 0x7FFFFFFF, OFFSET = 0 };
  };

  template <>
  struct WAVtraits<uint32_t>
  {
    enum { MAGIC = WAVE_FORMAT_UNKNOWN, MAX = 0xFFFFFFFF, OFFSET = 0x80000000 };
  };

  template <>
  struct WAVtraits<float>
  {
    enum { MAGIC = WAVE_FORMAT_IEEE_FLOAT, MAX = 1, OFFSET = 0 };
  };

  template <>
  struct WAVtraits<double>
  {
    enum { MAGIC = WAVE_FORMAT_UNKNOWN, MAX = 1, OFFSET = 0 };
  };

}

template <typename _Iter>
int WAVsave(const char *filename, int channels, int rate, _Iter begin, const _Iter end)
{
  typedef typename std::iterator_traits<_Iter>::value_type T;

  BOOST_STATIC_ASSERT(WAVtraits<T>::MAGIC != WAVE_FORMAT_UNKNOWN); /* if this fails _Iter provided a not supported sample type */

  WAV wav;
  wav.fmt.type=WAVtraits<T>::MAGIC;
  wav.fmt.channels=channels;
  wav.fmt.rate=rate;
  wav.fmt.bitsPerSample=sizeof(T)*8;
  wav.fmt.fix();

  doj::FILEp f(filename, "wb");
  if(!f)
    {
      fprintf(stderr, "could not open %s\n", filename);
      return -1;
    }
  if(fwrite(&wav, sizeof(wav), 1, f) != 1)
    return -1;

  while(begin!=end)
    {
      if(fwrite(&(*begin), sizeof(T), 1, f) != 1)
	return -1;
      ++begin;
    }

  wav.setlength(ftell(f));
  if(fseek(f, 0, SEEK_SET) != 0)
    return -1;
  if(fwrite(&wav, sizeof(wav), 1, f) != 1)
    return -1;

  return 0;
}

#include <memorymap.hpp>

template <typename T>
T* WAVload(const char *filename, int& channels, int& samplerate, unsigned& size, T& maxvalue, const char* &errmsg)
{
  BOOST_STATIC_ASSERT(WAVtraits<T>::MAX); /* if this fails T provided a not supported sample type */

  T *buf=0;

  WAV *w=reinterpret_cast<WAV*>(doj::memorymap(filename));
  if(!w)
    {
      errmsg="could not memorymap file";
      return 0;
    }

  if((unsigned)doj::memorymapsize(w) < sizeof(WAV))
    {
      errmsg="filesize is too small";
      goto WAVloadexit;
    }

  if(!w->iswav())
    {
      errmsg="filetype is not wav";
      goto WAVloadexit;
    }

  // some more checks
  if(w->fmt.type == WAVE_FORMAT_IEEE_FLOAT && w->fmt.bytesPerSample != sizeof(float))
    {
      errmsg="format float header error";
      goto WAVloadexit;
    }
  if(w->fmt.type == WAVE_FORMAT_PCM)
    switch(w->fmt.bitsPerSample)
      {
      case 8: case 12: case 16: case 32: break;	/* we can read those */
      case 24: errmsg="24bit integer not supported"; goto WAVloadexit;
      default: errmsg="unknown integer bit size"; goto WAVloadexit;
      }
  if(w->fmt.bitsPerSample < 8)
    {
      errmsg="bitsPerSample < 8";
      goto WAVloadexit;
    }

  channels=w->fmt.channels;
  samplerate=w->fmt.rate;
  size=w->data.length / (w->fmt.bitsPerSample/8);
  maxvalue=static_cast<T>(0);

  buf=new T[size];
  if(w->fmt.type == WAVE_FORMAT_IEEE_FLOAT)
    {
      float *src=reinterpret_cast<float*>(w->data.data());
      for(unsigned i=0; i!=size; ++i)
	{
	  buf[i]=static_cast<T>(src[i] * WAVtraits<T>::MAX + WAVtraits<T>::OFFSET);
	  if(buf[i]>maxvalue)
	    maxvalue=buf[i];
	}
    }
  if(w->fmt.type == WAVE_FORMAT_PCM)
    switch(w->fmt.bitsPerSample)
      {

#define CONV(type) \
  type *src=reinterpret_cast<type *>(w->data.data()); \
  for(unsigned i=0; i!=size; ++i) \
  { \
    buf[i]=(static_cast<T>(src[i]) - WAVtraits<type>::OFFSET) * WAVtraits<T>::MAX / WAVtraits<type>::MAX + WAVtraits<T>::OFFSET; \
    if(buf[i]>maxvalue) \
      maxvalue=buf[i]; \
  }

      case 8:
	{
	  CONV(uint8_t);
	  break;
	}
	break;
      case 12: case 16:
	{
	  CONV(int16_t);
	  break;
	}
	break;
      case 32:
	{
	  CONV(int32_t);
	  break;
	}
	break;
      default: assert(0); /* should not happen at this point */

#undef CONV
      }

 WAVloadexit:
  doj::memoryunmap(w);
  return buf;
}

#endif /* __cplusplus */

#endif /* WAV__H */
